# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2017, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# ******************************************************************

import numpy as np
import sys
import math
import warnings

from _snap_ml.utils import _is_mpi_enabled
_mpi_enabled = _is_mpi_enabled()

if sys.version_info[0] < 3:
    import libsnapmllocal2 as libsnapml
else:
    import _snap_ml.libsnapmllocal3 as libsnapml

class BoostingMachine:

    """
    BoostingMachine

    This class implements a boosting machine that can be used to construct an ensemble of decision trees.
    It can be used for both clasification and regression tasks.
    In constrast to other boosting frameworks, Snap ML's boosting machine dose not utilize a fixed maximal tree depth at each boosting iteration.
    Instead, the tree depth is sampled at each boosting iteration according to a discrete uniform distribution.
    The fit and predict functions accept numpy.ndarray data structures.

    Parameters
    ----------
    objective: string, optional, default : "mse"
        The training objective ("mse" or "logloss").
        "mse" is typically used for regression tasks, and "logloss" for classification tasks.

    num_round: integer, optional, default : 10
        The number of boosting iterations.

    min_max_depth: integer, optional, default : 1
        Minimum max_depth of a tree in the ensemble.

    max_max_depth: integer, optional, default : 6
        Maximum max_depth of a tree in the ensemble.

    n_threads: integer, optional, default : 1
        The number of CPU threads to use.

    random_state : integer, or None, optional, default : None
        If integer, random_state is the seed used by the random number generator.
        If None, the random number generator is the RandomState instance used by `np.random`.

    learning_rate: float, optional, default : 0.1
        Learning rate in the boosting algorithm.

    use_histograms: boolean, default : True
        Use histogram-based splits rather than exact splits.

    hist_nbins: int, default : 256
        Number of histogram bins.

    use_gpu: boolean, default : False
        Use GPU acceleration (only supported for histogram-based splits).

    gpu_id: int, default: 0
        Device ID of the GPU which will be used when GPU acceleration is enabled.

    verbose : boolean, default : False
        If True, it prints debugging information while training.
        Warning: this will increase the training time. For performance evaluation, use verbose=False.

    colsample_bytree : float, default : 1.0
        The fraction of features to be subsampled at each boosting iteration.
        The value range of this parameter is (0,1].

    subsample : float, default : 1.0
        The fraction of examples to be subsampled at each boosting iteration.
        The value range of this parameter is (0,1].

    parallel_by_example : boolean, default : False
        If True, computation of histogram bins will be parallelized by example (rather than by feature).
        This option may reduce training time for datatsets with a million or more examples.
        Only relevant for CPU-based training.

    lambda_l2 : float, default : 0.0
        L2-regularization parameter applied to tree weights.


    Attributes
    ----------

    """

    def __init__(self,  objective='mse',
                        num_round=10,
                        min_max_depth = 1,
                        max_max_depth = 6,
                        n_threads = 1,
                        random_state=None,
                        learning_rate = 0.1,
                        use_histograms=True,
                        hist_nbins=256,
                        use_gpu=False,
                        gpu_id=0,
                        verbose = False,
                        colsample_bytree = 1.0,
                        subsample = 1.0,
                        parallel_by_example=False,
                        lambda_l2 = 0.0):


        self.num_round_ = num_round
        if num_round <= 0:
            raise ValueError("Parameter num_round should be an integer >= 1.")

        self.objective_ = objective
        if objective not in ['mse', 'logloss']:
            raise ValueError("Objective can be 'mse' or 'logloss'")

        self.min_max_depth_ = min_max_depth
        if min_max_depth <= 0:
            raise ValueError("Parameter min_max_depth should be an integer >= 1.")

        self.max_max_depth_ = max_max_depth
        if max_max_depth <= 0:
            raise ValueError("Parameter max_max_depth should be an integer >= 1.")

        if max_max_depth < min_max_depth:
            raise ValueError("Parameter max_max_depth should be >= min_max_depth")

        self.learning_rate_ = learning_rate
        if learning_rate < 0.0:
            raise ValueError("Parameter learning rate should be positive.");

        self.random_state_ = random_state
        if (not isinstance(random_state, (int))) and (not random_state is None):
            raise ValueType("Parameter random_state: invalid value or type. Valid type is int or None.")

        self.n_threads_ = n_threads
        if n_threads < 0:
            raise ValueError("Parameter n_threads should be an integer >= 0.")

        self.use_histograms_ = use_histograms
        if not isinstance(use_histograms, (bool)):
            raise TypeError("Parameter use_histograms: invalid type. Valid type is bool {True, False}.")

        self.hist_nbins_ = hist_nbins
        if not isinstance(hist_nbins, (int)):
            raise TypeError("Parameter hist_nbins: invalid type. Valid type is int.")
        if hist_nbins < 1 or hist_nbins > 256:
            raise ValueError("Number of bins must be between 1 and 256")

        self.use_gpu_ = use_gpu
        if not isinstance(use_gpu, (bool)):
            raise TypeError("Parameter use_gpu: invalid type. Valid type is bool {True, False}.")

        if (use_gpu == True) and (use_histograms == False):
            raise ValueError("GPU acceleration can only be enabled if use_histograms=True")

        self.gpu_id_ = gpu_id
        if gpu_id < 0:
            raise ValueError("Parameter gpu_id should be an integer >= 0")

        self.verbose_ = verbose
        if not isinstance(verbose, (bool)):
            raise TypeError("Parameter verbose: invalid type. Valid type is bool {True, False}.")
 
        self.colsample_bytree_ = colsample_bytree
        if colsample_bytree <= 0 or colsample_bytree > 1:
            raise ValueError("Parameter colsample_bytree: invalid value. Valid values are in (0,1].")

        self.subsample_ = subsample
        if subsample <= 0 or subsample > 1:
            raise ValueError("Parameter subsample: invalid value. Valid values are in (0,1].")

        self.parallel_by_example_ = parallel_by_example
        if not isinstance(verbose, (bool)):
            raise TypeError("Parameter parallel_by_example invalid type. Valid type is bool {True, False}.")

        self.lambda_l2_ = lambda_l2
        if lambda_l2 < 0:
            raise ValueError("Parameter lambda should be >= 0")


    def get_params(self):

        """
        Get the values of the model parameters.

        Returns
        -------
        params : dict
        """

        params = {}
        params['num_round'] = self.num_round_
        params['objective'] = self.objective_
        params['min_max_depth'] = self.min_max_depth_
        params['max_max_depth'] = self.max_max_depth_
        params['random_state'] = self.random_state_
        params['n_threads'] = self.n_threads_
        params['learning_rate'] = self.learning_rate_
        params['use_histograms'] = self.use_histograms_
        params['hist_nbins'] = self.hist_nbins_
        params['use_gpu'] = self.use_gpu_
        params['gpu_id'] = self.gpu_id_
        params['verbose'] = self.verbose_
        params['colsample_bytree'] = self.colsample_bytree_
        params['subsample'] = self.subsample_
        params['parallel_by_example'] = self.parallel_by_example_
        params['lambda_l2'] = self.lambda_l2_

        return params

    def fit(self, X_train, y_train):

        """
        Fit the model according to the given train data.

        Parameters
        ----------
        X_train : dense matrix (ndarray)
            Train dataset

        y_train : array-like, shape = (n_samples,)
            The target vector corresponding to X_train.

        Returns
        -------
        self : object
        """

        # Random state
        if (self.random_state_ == None):
            # Not sure if this should be the random state
            random_state = np.random.get_state()[1][0]
        else:
            random_state = self.random_state_

        if type(X_train).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        num_ft = 0
        num_nz = 0
        indptr = np.array([])
        indices = np.array([])
        data = np.array([])
        labs = np.array([])

        num_ex = len(y_train)

        # Check train dataset format
        if (X_train.shape[0] != num_ex):
            raise ValueError("X_train.shape[0] must be equal to number of examples.")

        if (type(X_train).__name__ == "csr_matrix"):
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "csc_matrix"):
            X_train = X_train.tocsr()
            num_ft = X_train.shape[1]
        elif (type(X_train).__name__ == "ndarray"):
            if (X_train.shape[0] == X_train.shape[1]):
                print("[Warning] Each row in the input train dataset is assumed to be a training sample.")
                num_ft = X_train.shape[0]
            elif (X_train.shape[0] == num_ex):
                num_ft = X_train.shape[1]
            else:
                X_train = X_train.transpose().copy()
                num_ft = X_train.shape[1]

        if (type(X_train).__name__ == "csr_matrix" or type(X_train).__name__ == "csc_matrix"):
            num_nz    = X_train.nnz
            indptr    = X_train.indptr.astype(np.uint64)
            data      = X_train.data.astype(np.float32)
            indices   = X_train.indices.astype(np.uint32)
        elif (type(X_train).__name__ == "ndarray"):
            num_nz    = num_ex * num_ft
            data      = X_train.astype(np.float32)
            data      = np.ascontiguousarray(data)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise TypeError("X_train should be in csr_matrix, csc_matrix or ndarray format.")


        labs = y_train.astype(np.float32)


        out_fit = libsnapml.booster_fit(self.objective_,
                                        self.num_round_,
                                        self.min_max_depth_,
                                        self.max_max_depth_,
                                        self.colsample_bytree_,
                                        self.subsample_,
                                        random_state,
                                        self.learning_rate_,
                                        self.n_threads_,
                                        self.use_histograms_,
                                        self.hist_nbins_,
                                        self.use_gpu_,
                                        self.gpu_id_,
                                        self.verbose_,
                                        self.parallel_by_example_,
                                        self.lambda_l2_,
                                        num_ex,
                                        num_ft,
                                        num_nz,
                                        indptr,
                                        indices,
                                        data,
                                        labs)

        self.model_ = out_fit[0]
        training_metadata = out_fit[1]
        self.model_size_ = training_metadata['model_size']
        return None

    def predict(self, X):

        """
        Raw predictions

        If the training objective is 'mse' then it returns the predicted estimates.
        If the training objective is 'logloss' then it returns the predicted estimates before the logistic transformation (raw logits).

        Parameters
        ----------
        X : dense matrix (ndarray)
            Dataset used for predicting class estimates.

        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted estimates.
        """

        if type(X).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])

        num_ex = X.shape[0]
        num_ft = X.shape[1]

        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # flag that indicates if output probabilities are to be computed
        # 0 : get raw predictions
        # 1 : get output probabilities (only for predict proba)
        get_proba = 0

        # Generate predictions
        pred = libsnapml.booster_predict(num_ex,
                                         num_ft,
                                         indptr,
                                         indices,
                                         data,
                                         self.num_round_,
                                         self.learning_rate_,
                                         self.model_,
                                         self.model_size_,
                                         get_proba)

        pred = np.array(pred)

        if not np.all(np.isfinite(pred)):
            warnings.warn("Boosting diverged; Try using a smaller learning rate.")

        # handle case of divergence
        pred = np.nan_to_num(pred)

        return pred


    def predict_proba(self, X):

        """
        Output probabilities

        Use only if the training objective is 'logloss' (i.e., for binary classification problems).
        It returns the probabilities of each sample belonging to each class. 
        The probabilities are calculated using the logistic transformation. 

        Parameters
        ----------
        X : dense matrix (ndarray)
            Dataset used for predicting class estimates.

        Returns
        -------
        proba: array-like, shape = (n_samples, 2)
            Returns the predicted probabilities of each sample belonging to each class.
        """

        if type(X).__name__ != "ndarray":
            raise TypeError("Tree-based models in Snap ML only support numpy.ndarray")

        num_ex = 0
        num_ft = 0
        indptr = np.array([])
        indices =  np.array([])
        data = np.array([])

        num_ex = X.shape[0]
        num_ft = X.shape[1]

        if (type(X).__name__ == "csc_matrix"):
            X = X.tocsr()
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "csr_matrix"):
            indptr    = X.indptr.astype(np.uint64)
            data      = X.data.astype(np.float32)
            indices   = X.indices.astype(np.uint32)
        elif (type(X).__name__ == "ndarray"):
            data      = X.astype(np.float32)
            indptr    = np.array([])
            indices   = np.array([])
        else:
            raise ValueError("X should be in csc_matrix, csr_matrix or ndarray format.")

        pred = []

        # flag that indicates if output probabilities are to be computed
        # 0 : get raw predictions
        # 1 : get output probabilities (only for predict proba)
        get_proba = 1

        # Generate predictions
        pred = libsnapml.booster_predict(num_ex,
                                         num_ft,
                                         indptr,
                                         indices,
                                         data,
                                         self.num_round_,
                                         self.learning_rate_,
                                         self.model_,
                                         self.model_size_,
                                         get_proba)

        pred = np.array(pred)

        if not np.all(np.isfinite(pred)):
            warnings.warn("Boosting diverged; Try using a smaller learning rate.")

         # handle case of divergence
        pred = np.nan_to_num(pred)

        return pred

