# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import numpy as np
from pai4sk.metrics import mean_squared_errorSklearn
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
_mpi_enabled = _is_mpi_enabled()

def mean_squared_error(y_true, y_pred,
                       sample_weight=None,
                       multioutput='uniform_average', squared=True):
    """Mean squared error regression loss

    For SnapML solver this supports both local and distributed(MPI) method of execution.

    Read more in the :ref:`User Guide <mean_squared_error>`.

    Parameters
    ----------
    y_true : array-like of shape = (n_samples) or (n_samples, n_outputs)
        Ground truth (correct) target values.
        It also accepts SnapML data partition, which includes the correct labels.

    y_pred : array-like of shape = (n_samples) or (n_samples, n_outputs)
        Estimated target values.

    sample_weight : array-like of shape = (n_samples), optional
        Sample weights.

    multioutput : string in ['raw_values', 'uniform_average']
        or array-like of shape (n_outputs)
        Defines aggregating of multiple output values.
        Array-like value defines weights used to average errors.

        'raw_values' :
            Returns a full set of errors in case of multioutput input.

        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.

    squared : boolean value, optional (default = True)
        If True returns MSE value, if False returns RMSE value.

    Returns
    -------
    loss : float or ndarray of floats
        A non-negative floating point value (the best value is 0.0), or an
        array of floating point values, one for each individual target.

    Examples
    --------
    >>> from pai4sk.metrics import mean_squared_error
    >>> y_true = [3, -0.5, 2, 7]
    >>> y_pred = [2.5, 0.0, 2, 8]
    >>> mean_squared_error(y_true, y_pred)
    0.375
    >>> y_true = [3, -0.5, 2, 7]
    >>> y_pred = [2.5, 0.0, 2, 8]
    >>> mean_squared_error(y_true, y_pred, squared=False)
    0.612...
    >>> y_true = [[0.5, 1],[-1, 1],[7, -6]]
    >>> y_pred = [[0, 2],[-1, 2],[8, -5]]
    >>> mean_squared_error(y_true, y_pred)
    0.708...
    >>> mean_squared_error(y_true, y_pred, multioutput='raw_values')
    array([0.41666667, 1.        ])
    >>> mean_squared_error(y_true, y_pred, multioutput=[0.3, 0.7])
    0.825...

    """
    # Use snap-ml implementation only for partitoned data-type or MPI
    if _is_data_snapml_partition_type(y_true) or _mpi_enabled:
        # Check if any args apart from the ones supported by snap-ml mean_squared_error is provided
        if( (sample_weight == None) and
            (multioutput == 'uniform_average') ):
            #Input arguments are supported by snapML mean_squared_error
            from pai4sk import Metrics
            print("[Info] PowerAI: Using snapML optimized mean_squared_error") 
            mse = Metrics.mean_squared_error(y_true, y_pred)
            return mse if squared else np.sqrt(mse)
        else:
            raise ValueError("PowerAI: snapML optimized mean_squared_error does not support "+
                             "the values for parameters : sample_weight, multioutput. "+
                             "Do not set these parameters.")
    else:
        return mean_squared_errorSklearn(y_true, y_pred, sample_weight, multioutput, squared)
