# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import time
import numpy as np
import sys
import argparse

import cupy
from numba import cuda

defaultPath = "."

CLI=argparse.ArgumentParser()
CLI.add_argument(
    "--data_path",
    type=str,
    default=defaultPath
)

args = CLI.parse_args()
t0 = time.time()

# Load train data as cupy array
X_train_cupy = cupy.load(args.data_path + "/data/epsilon.X_train.npy").astype(np.float32)
y_train_cupy = cupy.load(args.data_path + "/data/epsilon.y_train.npy").astype(np.float32)

# Create DeviceNDArray from cupy array for SnapML. No copying of the data is done.
X_train_dnd = cuda.as_cuda_array(X_train_cupy)
y_train_dnd = cuda.as_cuda_array(y_train_cupy)

# Create DeviceNDArray F order from cupy array for cuML. Another copy of data is made.
X_train_dnd_F = cuda.as_cuda_array(cupy.copy(X_train_cupy, order='F'))
y_train_dnd_F = cuda.as_cuda_array(cupy.copy(y_train_cupy, order='F'))

# Load test data as numpy array
X_test_npy = np.load(args.data_path + "/data/epsilon.X_test.npy").astype(np.float32)
y_test_npy = np.load(args.data_path + "/data/epsilon.y_test.npy").astype(np.float32)

print("Data load time (s):  {0:.2f}".format(time.time()-t0))

penalty = 0.5

# Create a SVM instance for SnapML with pai4sk explicit call
from pai4sk import SupportVectorMachine
snap_svm = SupportVectorMachine(use_gpu=True, regularizer=1/penalty)

# SnapML - Training
t0 = time.time()
snap_svm.fit(X_train_dnd, y_train_dnd)
print("[SnapML] Training time (s): {0:.2f}".format(time.time()-t0))

# SnapML - predict
pred = snap_svm.predict(X_test_npy)

# SnapML - Evaluate accuracy on test set
from sklearn.metrics import accuracy_score
acc_snap = accuracy_score(y_test_npy, pred)
print("[SnapML] Accuracy : {0:.3f}".format(acc_snap))

# SnapML - decision_function
pred = snap_svm.decision_function(X_test_npy)

# SnapML - hinge_loss on test set
from sklearn.metrics import hinge_loss
loss_snap = hinge_loss(y_test_npy, pred)
print("[SnapML] Hinge-loss : {0:.3f}".format(loss_snap))

# Import cuML's SVC
from cuml.svm import SVC
cuml_svc = SVC(C=0.5, kernel='linear')

# cuML - Training
t0 = time.time()
cuml_svc.fit(X_train_dnd_F, y_train_dnd_F)
print("[cuML] Training time (s):  {0:.2f}".format(time.time()-t0))

# cuML - predict
pred = cuml_svc.predict(X_test_npy)

# cuML - Evaluate accuracy on test set
from sklearn.metrics import accuracy_score
acc = accuracy_score(y_test_npy, pred.to_array())
print("[cuML] Accuracy: {0:.3f}".format(acc))

# cuML - decision_function
pred = cuml_svc.decision_function(X_test_npy)

# cuML - hinge_loss on test set
from sklearn.metrics import hinge_loss
loss = hinge_loss(y_test_npy, pred.to_array())
print("[cuML] Hinge loss: {0:.3f}".format(loss))

del(cuml_svc)
